// copyright (2019.) Marko Čupić - Fakultet elektrotehnike i računarstva, Sveučilište u Zagrebu; nije dozvoljena uporaba u vlastitim projektima, prerada ili daljna distribucija bez dozvole autora

function AutomatonState(name) {
  this.stateName = name;
  this.center = new vector([0,0]);
  this.radius = 30;
  this.outputs = null;
  this.isSelected = false;
  this.isActive = false;
  this.isStart = false;
}
AutomatonState.importStateDef = function(def) {
  let st = new AutomatonState(def.stateName);
  st.center = new vector(def.center.elems);
  st.radius = def.radius;
  st.outputs = def.outputs;
  st.isSelected = def.isSelected;
  st.isActive = def.isActive;
  st.isStart = def.isStart;
  return st;
}

function AutomatonTransition(stateName1, stateName2) {
  this.stateName1 = stateName1;
  this.stateName2 = stateName2;
  this.hFactor = 0.5;
  this.vFactor = 0.5;
  this.angle = Math.PI;
  this.fov = Math.PI/4;
  this.polygonVisible = false;
  this.inputFilter = [];
}

AutomatonTransition.importTransitionDef = function(def) {
  let tr = new AutomatonTransition(def.stateName1, def.stateName2);
  tr.hFactor = def.hFactor;
  tr.vFactor = def.vFactor;
  tr.angle = def.angle;
  tr.fov = def.fov;
  tr.polygonVisible = def.polygonVisible;
  tr.inputFilter = def.inputFilter;
  if(tr.inputFilter) {
    for(let i = 0; i < tr.inputFilter.length; i++) {
      let f = tr.inputFilter[i];
      for(let j = 0; j < f.length; j++) {
        if(f[j].isATWildcard) f[j]=tr.WILDCARD;
      }
    }
  }
  return tr;
}

AutomatonTransition.prototype.WILDCARD = {isATWildcard: true};
AutomatonTransition.prototype.filtersAsString = function() {
  let res = "";
  for(let i = 0; i < this.inputFilter.length; i++) {
    if(i>0) res += ";";
    for(let j = 0; j < this.inputFilter[i].length; j++) {
      if(j>0) res += ",";
      if(this.inputFilter[i][j]==this.WILDCARD) res += "*"; else res += this.inputFilter[i][j];
    }
  }
  return res;
}

// vraća 1 ako je f1 općenitiji od f2, 0 ako su jednako općeniti, -1 inače
AutomatonTransition.prototype.compareTwoFilters = function(f1,f2) {
  let win = 0;
  for(let i = 0; i < f1.length; i++) {
    if(f1[i]==this.WILDCARD) {
      if(f2[i]!=this.WILDCARD) {
        win++;
      }
    } else if(f1[i]!=f2[i]) return -1;
  }
  return win==0 ? 0 : 1;
}

function AutomatonModel() {
  this.states = {};
  this.stateNames = [];
  this.transitions = [];
  this.inputs = []; // objekti: {name: 'xxx', values: ['..','..','..']}
  this.outputs = []; // objekti: {name: 'xxx', values: ['..','..','..']}
  this.currentInput = [];
  this.currentOutput = [];
}

AutomatonModel.importModelDef = function(def) {
  let mo = new AutomatonModel();
  if(def.stateNames) {
    for(let i = 0; i < def.stateNames.length; i++) {
      let st = def.states["st_"+def.stateNames[i]];
      let stClone = AutomatonState.importStateDef(st);
      mo.stateNames.push(stClone.stateName);
      mo.states["st_"+stClone.stateName] = stClone;
    }
  }
  if(def.inputs) {
    for(let i = 0; i < def.inputs.length; i++) {
      mo.inputs.push(def.inputs[i]);
    }
  }
  if(def.outputs) {
    for(let i = 0; i < def.outputs.length; i++) {
      mo.outputs.push(def.outputs[i]);
    }
  }
  if(def.transitions) {
    for(let i = 0; i < def.transitions.length; i++) {
      mo.transitions.push(AutomatonTransition.importTransitionDef(def.transitions[i]));
    }
  }
  return mo;
}

AutomatonModel.prototype.findActiveState = function() {
  for(let i = 0; i < this.stateNames.length; i++) {
    let st = this.states["st_"+this.stateNames[i]];
    if(st.isActive) return st;
  }
  return null;
}

AutomatonModel.prototype.findStartState = function() {
  for(let i = 0; i < this.stateNames.length; i++) {
    let st = this.states["st_"+this.stateNames[i]];
    if(st.isStart) return st;
  }
  return null;
}

AutomatonModel.prototype.setActiveState = function(stateName) {
  if(this.stateNames.indexOf(stateName)==-1) throw "State with given name does not exist.";
  let sst = this.findActiveState();
  if(sst!=null) sst.isActive=false;
  this.states["st_"+stateName].isActive = true;
}

AutomatonModel.prototype.checkAllDefined = function() {
  let inputs = this.inputs;
  let me = this;
  let currentInputs = [];
  let currentState = null;
  let messages = [];
  let result = {messages: messages};
  let missingTransCounter = 0;
  let missingTransLimit = 10;
  function setCurIndRecursive(ind) {
    if(ind<inputs.length) {
      for(let i = 0; i < inputs[ind].values.length; i++) {
        currentInputs[ind] = inputs[ind].values[i];
        setCurIndRecursive(ind+1);
      }
      return;
    }
    // Nađi tranziciju koja pali za zadano stanje i pobudu:
    let tr = me.findTransitionFor(currentState,currentInputs);
    if(tr==null) {
      missingTransCounter++;
      if(missingTransCounter<=missingTransLimit) {
        if(currentInputs.length==0) {
          messages.push("Missing transition from state "+currentState.stateName+".");
        } else {
          messages.push("Missing transition from state "+currentState.stateName+" and inputs "+currentInputs.join(",")+".");
        }
      }
    }
  }

  // Provjeri da ima barem jedan izlaz
  if(this.outputs.length==0) {messages.push("Automaton must have at least one output defined!"); return result; }

  // Provjeri da ima zadano početno stanje
  let hasStartState = false;
  for(let i = 0; i < this.stateNames.length; i++) {
    if(this.states["st_"+this.stateNames[i]].isStart) { hasStartState=true; break; }
  }
  if(!hasStartState) messages.push("Automaton must have start state!");

  // Provjeri da je u svakom stanju za svaku kombinaciju ulaza definiran prijelaz
  for(let i = 0; i < this.stateNames.length; i++) {
    currentState = this.states["st_"+this.stateNames[i]];
    setCurIndRecursive(0);
  }
  if(missingTransCounter>missingTransLimit) {
    messages.push("There are also other missing transitions ("+(missingTransCounter-missingTransLimit)+"); details were provided for first "+missingTransLimit+".");
  }
  return result;
}

AutomatonModel.prototype.addInput = function(i) {
  i.name = i.name.trim();
  if(i.name==="") throw "Name was not provided."
  if(this.hasInputOrOutput(i.name)) throw "Input or Output with this name already exists.";
  if(!i.values || i.values.length==0) throw "Input must have at least one defined legal value.";
  for(let idx = 0; idx < i.values.length; idx++) {
    if(i.values.indexOf(i.values[idx])!=idx) throw "Legal value '"+i.values[idx]+"' defined multiple times.";
  }
  this.inputs.push(i);
}

AutomatonModel.prototype.addOutput = function(o) {
  o.name = o.name.trim();
  if(o.name==="") throw "Name was not provided."
  if(this.hasInputOrOutput(o.name)) throw "Input or Output with this name already exists.";
  if(!o.values || o.values.length==0) throw "Output must have at least one defined legal value.";
  for(let idx = 0; idx < o.values.length; idx++) {
    if(o.values.indexOf(o.values[idx])!=idx) throw "Legal value '"+o.values[idx]+"' defined multiple times.";
  }
  this.outputs.push(o);
}

AutomatonModel.prototype.hasInputOrOutput = function(name) {
  for(let i = 0; i < this.inputs.length; i++) if(this.inputs[i].name==name) return true;
  for(let i = 0; i < this.outputs.length; i++) if(this.outputs[i].name==name) return true;
  return false;
}

AutomatonModel.prototype.addState = function(s) {
  if(this.stateNames.indexOf(s.stateName)!=-1) throw "State with same name already exists.";
  this.stateNames.push(s.stateName);
  this.states["st_"+s.stateName] = s;
}

AutomatonModel.prototype.setStartState = function(stateName) {
  if(this.stateNames.indexOf(stateName)==-1) throw "State with given name does not exist.";
  let sst = this.findStartState();
  if(sst!=null) sst.isStart=false;
  this.states["st_"+stateName].isStart = true;
}

AutomatonModel.prototype.addStateWithChecks = function(s) {
  if(this.stateNames.indexOf(s.stateName)!=-1) throw "State with same name already exists.";
  if(this.outputs.length>0) {
    if(s.outputs==null) throw "State has no outputs defined, but automaton has outputs!";
    if(s.outputs.length != this.outputs.length) throw "State outputs and automaton outputs have different number of components.";
    for(let i = 0; i < this.outputs.length; i++) {
      let odef = this.outputs[i];
      let sval = s.outputs[i];
      if(odef.values.indexOf(sval)==-1) throw "Value '"+sval+"' is invalid for output "+odef.name+".";
    }
  }
  this.stateNames.push(s.stateName);
  this.states["st_"+s.stateName] = s;
}

AutomatonModel.prototype.getState = function(n) {
  return this.states["st_"+n];
}

AutomatonModel.prototype.addTransition = function(t) {
  this.transitions.push(t);
}

AutomatonModel.prototype.findTransitionFor = function(state,inputs) {
  let candidateTransitions = this.transitions.filter(t => t.stateName1==state.stateName);
  let selectedTransition = null;
  let strength = 0;

  for(let i = 0; i < candidateTransitions.length; i++) {
    let ct = candidateTransitions[i];
    for(let j = 0; j < ct.inputFilter.length; j++) {
      let f = ct.inputFilter[j];
      let fstrength = 0;
      for(let k = 0; k < f.length; k++) {
        if(f[k]==ct.WILDCARD) continue;
        if(f[k]!=inputs[k]) { fstrength = -1; break; }
        fstrength++;
      }
      if(fstrength==-1) continue;
      if(selectedTransition==null || fstrength>strength) {
        strength = fstrength;
        selectedTransition = ct;
      }
    }
  }
  return selectedTransition;
}

AutomatonModel.prototype.addTransitionWithCheck = function(tr) {
  if(this.stateNames.indexOf(tr.stateName1)==-1) throw "State with name '"+tr.stateName1+"' does not exists.";
  if(this.stateNames.indexOf(tr.stateName2)==-1) throw "State with name '"+tr.stateName2+"' does not exists.";
  let candidateTransitions = this.transitions.filter(t => t.stateName1==tr.stateName1);
  let f1 = tr.inputFilter[0];
  let left = [];
  for(let i = 0; i < candidateTransitions.length; i++) {
    let ct = candidateTransitions[i];
    for(let j = 0; j < ct.inputFilter.length; j++) {
      let f2 = ct.inputFilter[j];
      let cres = tr.compareTwoFilters(f2,f1);
      if(cres==1 && tr.stateName2!=ct.stateName2) continue;
      if(cres==1 && tr.stateName2==ct.stateName2) throw "Abort: applicable transition already exists.";
      if(cres==0 && tr.stateName2==ct.stateName2) throw "Abort: same transition already exists.";
      if(cres==0 && tr.stateName2!=ct.stateName2) throw "Abort: conflicting transition already exists.";
      if(tr.stateName2==ct.stateName2 && left.indexOf(ct)==-1) left.push(ct);
    }
  }
  if(left.length>0) {
    let ct = left[0];
    let arr = [];
    for(let i = 0; i < ct.inputFilter.length; i++) {
      let cres = tr.compareTwoFilters(f1,ct.inputFilter[i]); //TODO..
      if(cres < 0) arr.push(ct.inputFilter[i]);
    }
    arr.push(f1);
    ct.inputFilter.length = 0;
    for(let i = 0; i < arr.length; i++) {
      ct.inputFilter.push(arr[i]);
    }
  } else {
    tr.inputFilter.length = 1;
    this.transitions.push(tr);
  }
}

function AutomatonDisplay(canvasID, width, height, model) {
  this.width = width;
  this.height = height;
  this.canvas = document.getElementById(canvasID);
  this.context = this.canvas.getContext("2d");
  this.model = model;
  var regions = [];
  var ctx = this.context;
  let matrixB = new matrix([[-1,3,-3,1],[3,-6,3,0],[-3,3,0,0],[1,0,0,0]]);

  function drawCross(ctx,pt) {
    let crossHalf = 5;
    ctx.strokeStyle = "#00FF00";
    ctx.beginPath();
    ctx.moveTo(pt.elems[0]-crossHalf, pt.elems[1]-crossHalf);
    ctx.lineTo(pt.elems[0]+crossHalf, pt.elems[1]+crossHalf);
    ctx.stroke();
    ctx.beginPath();
    ctx.moveTo(pt.elems[0]-crossHalf, pt.elems[1]+crossHalf);
    ctx.lineTo(pt.elems[0]+crossHalf, pt.elems[1]-crossHalf);
    ctx.stroke();
  }

  function drawTransitionLink(ctx,t,r0,r1,r2,r3) {
    ctx.strokeStyle = "#0000FF";
    ctx.beginPath();
    ctx.moveTo(r0.elems[0], r0.elems[1]);
    ctx.bezierCurveTo(r1.elems[0], r1.elems[1], r2.elems[0], r2.elems[1], r3.elems[0], r3.elems[1]);
    ctx.stroke();

    let r3r2 = r3.sub(r2).normalize();
    let vert = r3r2.orthogonal2D();
    let baseLen = 10;
    let baseWidth = 5;
    let p1 = r3.sub(r3r2.scalarMultiply(baseLen)).add(vert.scalarMultiply(baseWidth));
    let p2 = r3.sub(r3r2.scalarMultiply(baseLen)).add(vert.scalarMultiply(-baseWidth));

    ctx.beginPath();
    ctx.moveTo(p1.elems[0], p1.elems[1]);
    ctx.lineTo(r3.elems[0], r3.elems[1]);
    ctx.stroke();
    ctx.beginPath();
    ctx.moveTo(p2.elems[0], p2.elems[1]);
    ctx.lineTo(r3.elems[0], r3.elems[1]);
    ctx.stroke();
  }

  function findClosestShape(regions, pt) {
    let closest = null;
    let distance = 0;
    let point = new vector([pt.x, pt.y]);
    for(let i = 0; i < regions.length; i++) {
      let r = regions[i];
      if(r.shape=="circle") {
        let d = (new vector([pt.x-r.center.elems[0],pt.y-r.center.elems[1]])).norm();
        if(d < r.radius) { d = 0; } else { d = d-r.radius; }
        if(closest==null || d<distance) { closest = r; distance = d; }
      } else if(r.shape=="point") {
        let d = point.sub(r.point).norm();
        if(closest==null || d<distance) { closest = r; distance = d; }
      } else if(r.shape=="polyline") {
        let v1 = new vector(r.points[0]);
        //console.log("gledam transition "+r.transition.stateName1+" na "+r.transition.stateName2);
        let d = 0;
        for(let j = 1; j < r.points.length; j++) {
          let v2 = new vector(r.points[j]);
          let p1p = v1.sub(point);
          let p2p1 = v2.sub(v1);
          let t = - p1p.scalarProduct(p2p1) / p2p1.scalarProduct(p2p1);
          if(t<0) {
            let dm = point.sub(v1).norm();
            if(j==1 || dm<d) { d = dm; }
          } else if(t>1) {
            let dm = point.sub(v2).norm();
            if(j==1 || dm<d) { d = dm; }
          } else {
            let dm = point.sub(v1.add(v2.sub(v1).scalarMultiply(t))).norm();
            if(j==1 || dm<d) { d = dm; }
          }
          v1 = v2;
        }
        //console.log("  ==> distanca je "+d);
        if(closest==null || d<distance) { closest = r; distance = d; }
      }
    }
    if(closest==null || distance>10) return null;
    return closest;
  }

  let pressState = {};
  pressState.regions = [];
  pressState.pressed = false;
  pressState.activeTransition = null;
  pressState.activeState = null;
  pressState.processMouseDown = function(me, pt) {
    pressState.pressed = false;
    pressState.region = null;
    let r = findClosestShape(pressState.regions, pt);
    if(r==null) {
      if(pressState.activeTransition!=null) { pressState.activeTransition.polygonVisible = false; pressState.activeTransition = null; }
      if(pressState.activeState!=null) { pressState.activeState.isSelected = false; pressState.activeState = null; }
      me.paint();
      return;
    }
    if(r.kind=="transition") {
      if(pressState.activeState!=null) { pressState.activeState.isSelected = false; pressState.activeState = null; }
      if(pressState.activeTransition!=null) pressState.activeTransition.polygonVisible = false;
      r.transition.polygonVisible = true;
      pressState.activeTransition = r.transition;
      me.paint();
      return;
    }
    if(r.kind=="state") {
      if(pressState.activeState!=null) { pressState.activeState.isSelected = false; }
      pressState.activeState = r.state;
      pressState.activeState.isSelected = true;
      if(pressState.activeTransition!=null) pressState.activeTransition.polygonVisible = false;
      pressState.cxy = [pt.x, pt.y];
      pressState.pressed = true;
      pressState.region = r;
      me.paint();
    }
    if(r.kind=="transitionVertex") {
      pressState.cxy = [pt.x, pt.y];
      pressState.pressed = true;
      pressState.region = r;
      me.paint();
    }
  }
  pressState.processMouseUp = function(me, pt) {
    if(!pressState.pressed) return;
    pressState.pressed = false;
  }
  pressState.processMouseMove = function(me, pt) {
    if(!pressState.pressed) return;
    if(pressState.region==null) return;
    if(pressState.region.kind=="state") {
      let delta = [pt.x-pressState.cxy[0],pt.y-pressState.cxy[1]];
      pressState.region.state.center.elems[0] += delta[0];
      pressState.region.state.center.elems[1] += delta[1];
      pressState.cxy = [pt.x, pt.y];
    } else if(pressState.region.kind=="transitionVertex") {
      let st1 = me.model.getState(pressState.region.transition.stateName1);
      let st2 = me.model.getState(pressState.region.transition.stateName2);
      if(st1!=st2) {
        let v1 = st1.center;
        let v2 = st2.center;
        let p1p = v1.sub(new vector([pt.x,pt.y]));
        let p2p1 = v2.sub(v1);
        let t = - p1p.scalarProduct(p2p1) / p2p1.scalarProduct(p2p1);
        pressState.region.transition.hFactor = t;

        let c1 = st1.center;
        let c2 = st2.center;
        let c2c1 = c2.sub(c1);
        let pv = c1.add(c2c1.scalarMultiply(t));
        let p = new vector([pt.x,pt.y]);
        let f = p.sub(pv).norm() / c2c1.norm();
        let v = pv.add(c2c1.orthogonal2D().normalize().scalarMultiply(c2c1.norm()*f));
        if(p.sub(pv).scalarProduct(v.sub(pv))<0) f = -f;
        pressState.region.transition.vFactor = f;
      } else {
        let c = st1.center;
        let v = new vector([pt.x,pt.y]);
        let vc = v.sub(c);
        let angle = Math.atan2(vc.elems[1], vc.elems[0]);
        let vFactor = vc.norm()/st1.radius-3;
        pressState.region.transition.vFactor = vFactor;
        pressState.region.transition.angle = angle;
      }
    }
    me.paint();
  }

  function generateArcPoints(r0,r1,r2,r3) {
    const n = 10;
    let tarr = [0,0,0,1];
    let tvec = new vector(tarr);
    let r = new matrix([r0.elems, r1.elems, r2.elems, r3.elems]);
    let br = matrixB.mul(r);
    let res = [];
    for(let i = 0; i < n; i++) {
      let t = i/(n-1.0);
      tarr[2] = t;
      tarr[1] = t*t;
      tarr[0] = tarr[1]*t;
      let pt = tvec.rightMultiplyBy(br);
      //console.log("pt: "+pt.elems[0]+", "+pt.elems[1]);
      res.push(pt.elems);
    }
    return res;
  }

  function generateHalfPoint(r0,r1,r2,r3) {
    let tarr = [0,0,0,1];
    let tvec = new vector(tarr);
    let tarr2 = [0,0,1,0];
    let tvec2 = new vector(tarr2);
    let r = new matrix([r0.elems, r1.elems, r2.elems, r3.elems]);
    let br = matrixB.mul(r);
    let t = 0.5;
    tarr[2] = t;
    tarr[1] = t*t;
    tarr[0] = tarr[1]*t;
    let pt = tvec.rightMultiplyBy(br);
    tarr2[1] = 2*t;
    tarr2[0] = 3*t*t;
    let ptt = tvec2.rightMultiplyBy(br);
    return [pt,ptt];
  }

  this.paint = function() {
    this.context.clearRect(0, 0, this.canvas.width, this.canvas.height);
    this.context.fillStyle = "#ffffee";
    this.context.strokeStyle = "#aaaa00";

    let tangentFactor = 0.75;

    regions.length = 0;
    pressState.regions = [];

    if(this.model.inputs.length!=0 || this.model.outputs.length!=0) {
      let str = "";
      for(let i = 0; i < this.model.inputs.length; i++) {
        if(i>0) str+=",";
        str += this.model.inputs[i].name;
      }
      str+="/";
      for(let i = 0; i < this.model.outputs.length; i++) {
        if(i>0) str+=",";
        str += this.model.outputs[i].name;
      }

      this.context.fillStyle = "#000000";
      this.context.font = 'bold 10px sans';
      this.context.textAlign = 'left';      
      this.context.textBaseline='top';
      this.context.fillText(str, 0, 0);
    }

    for(let i = 0; i < this.model.stateNames.length; i++) {
      let st = this.model.getState(this.model.stateNames[i]);
      this.context.strokeStyle = "#0000FF";
      this.context.fillStyle = st.isActive ? "#FF0000" : "#FFFF00";
      this.context.lineWidth = st.isSelected ? 5 : 1;
      this.context.beginPath();
      this.context.arc(st.center.elems[0],st.center.elems[1],st.radius,0*Math.PI,2*Math.PI);
      //console.log(st.center.elems[0]+", "+st.center.elems[1]+", "+st.radius);
      //this.context.closePath();
      this.context.fill();
      this.context.stroke();

      if(st.isStart) {
        this.context.lineWidth = 1;
        this.context.beginPath();
        this.context.arc(st.center.elems[0],st.center.elems[1],st.radius-(st.isSelected ? 5 : 3),0*Math.PI,2*Math.PI);
        this.context.stroke();
      }

      this.context.fillStyle = "#000000";
      this.context.font = 'bold 12px sans';
      this.context.textAlign = 'center';      
      this.context.textBaseline='bottom';
      this.context.fillText(st.stateName, st.center.elems[0], st.center.elems[1]-4);
      if(st.outputs != null && st.outputs.length>0) {
        this.context.font = '10px sans';
        this.context.textBaseline='top';
        this.context.fillText("["+st.outputs.join(",")+"]", st.center.elems[0], st.center.elems[1]+4);
      }

      pressState.regions.push({shape: "circle", kind: "state", center: st.center, radius: st.radius, state: st});
    }
    this.context.lineWidth = 1;

    for(let i = 0; i < this.model.transitions.length; i++) {
      let tr = this.model.transitions[i];
      let st1 = this.model.getState(tr.stateName1);
      let st2 = this.model.getState(tr.stateName2);

      let r0 = null;
      let r3 = null;
      let r1 = null;
      let r2 = null;

      if(st1 != st2) {
        let c1 = st1.center;
        let c2 = st2.center;
        let c2c1 = c2.sub(c1);
        let v = c1.add(c2c1.scalarMultiply(tr.hFactor)).add(c2c1.orthogonal2D().normalize().scalarMultiply(c2c1.norm()*tr.vFactor));

        let vc1 = v.sub(c1);
        let vc2 = v.sub(c2);
        r0 = c1.add(vc1.scalarMultiply(st1.radius/vc1.norm()));
        r3 = c2.add(vc2.scalarMultiply(st2.radius/vc2.norm()));
        r1 = v.scalarMultiply(tangentFactor).add(r0.scalarMultiply(1-tangentFactor));
        r2 = v.scalarMultiply(tangentFactor).add(r3.scalarMultiply(1-tangentFactor));

        if(tr.polygonVisible) {
          this.context.setLineDash([3,2]);
          this.context.beginPath();
          this.context.moveTo(r0.elems[0], r0.elems[1]);
          this.context.lineTo(v.elems[0], v.elems[1]);
          this.context.lineTo(r3.elems[0], r3.elems[1]);
          this.context.stroke();
          this.context.setLineDash([]);
          pressState.regions.push({shape: "point", kind: "transitionVertex", point: v, transition: tr});
        }
      } else {
        let c = st1.center;
        r0 = c.add(new vector([st1.radius*Math.cos(tr.angle-tr.fov/2),st1.radius*Math.sin(tr.angle-tr.fov/2)]));
        r3 = c.add(new vector([st1.radius*Math.cos(tr.angle+tr.fov/2),st1.radius*Math.sin(tr.angle+tr.fov/2)]));
        r1 = r0.add(r0.sub(c).scalarMultiply(2+tr.vFactor));
        r2 = r3.add(r3.sub(c).scalarMultiply(2+tr.vFactor));
        let factor = 3+tr.vFactor;
        let v = c.add((new vector([st1.radius*Math.cos(tr.angle),st1.radius*Math.sin(tr.angle)])).scalarMultiply(factor));
        if(tr.polygonVisible) {
          this.context.setLineDash([3,2]);
          this.context.beginPath();
          this.context.moveTo(r0.elems[0], r0.elems[1]);
          this.context.lineTo(r1.elems[0], r1.elems[1]);
          this.context.lineTo(v.elems[0], v.elems[1]);
          this.context.lineTo(r2.elems[0], r2.elems[1]);
          this.context.lineTo(r3.elems[0], r3.elems[1]);
          this.context.stroke();
          this.context.setLineDash([]);
          pressState.regions.push({shape: "point", kind: "transitionVertex", point: v, transition: tr});
        }
      }

      //drawCross(this.context,r0);
      //drawCross(this.context,r1);
      //drawCross(this.context,r2);
      //drawCross(this.context,r3);

      drawTransitionLink(this.context,tr,r0,r1,r2,r3);
      let text = tr.filtersAsString();
      if(text!="") {
        const factor = 3;
        let hphpp = generateHalfPoint(r0,r1,r2,r3);
        let hp = hphpp[0];
        let hpp = hphpp[1];
        this.context.fillStyle = "#000000";
        this.context.font = '10px sans';
        if(Math.abs(hpp.elems[0]) > factor*Math.abs(hpp.elems[1])) {
          // horizontalni smještaj
          this.context.textAlign = 'center';      
          this.context.textBaseline='bottom';
          this.context.fillText(text, hp.elems[0], hp.elems[1]);
        } else if(Math.abs(hpp.elems[1]) > factor*Math.abs(hpp.elems[0])) {
          // vertikalni smještaj
          this.context.textAlign = 'left';      
          this.context.textBaseline='middle';
          this.context.fillText(text, hp.elems[0]+4, hp.elems[1]);
        } else {
          // ukoso
          if(Math.sign(hpp.elems[0])==Math.sign(hpp.elems[1])) {
            this.context.textAlign = 'left';      
            this.context.textBaseline='bottom';
            this.context.fillText(text, hp.elems[0]+4, hp.elems[1]);
          } else {
            this.context.textAlign = 'left';      
            this.context.textBaseline='top';
            this.context.fillText(text, hp.elems[0]+4, hp.elems[1]);
          }
        }
      }
      pressState.regions.push({shape: "polyline", kind: "transition", points: generateArcPoints(r0,r1,r2,r3), transition: tr});
/*
      let pts = generateArcPoints(r0,r1,r2,r3);
      this.context.beginPath();
      this.context.moveTo(pts[0][0], pts[0][1]);
      for(let k = 0; k < pts.length; k++) {
        this.context.lineTo(pts[k][0], pts[k][1]);
      }
      this.context.stroke();
     */ 
    }

    //regions.push({x1: xs+dispX, y1: ys+dispY, x2: xe+dispX, y2: ye+dispY, x: x, y: y});

    //this.context.fillStyle = this.pixels[y][x]==0 ? this.backgroundColor : this.foregroundColor;
    //this.context.fillRect(Math.round(xs)+dispX,Math.round(ys)+dispY,Math.round(xe-xs),Math.round(ye-ys));
    //this.context.strokeStyle = this.gridColor;
    //this.context.beginPath();
    //this.context.moveTo(dispX+0,dispY+yline);
    //this.context.lineTo(dispX+dispW, dispY+yline);
    //this.context.stroke();
    //this.context.fillStyle = "#000000";
    //this.context.font = '10px sans';
    //this.context.textAlign = 'right';      
    //this.context.textBaseline='middle';
    //this.context.fillText(y, marginHorizontal-3, this.mathematical ? yline : yline + marginVertical);
  }

  var me = this;

  this.canvas.addEventListener('mousedown', function(evt) {
    var rect = me.canvas.getBoundingClientRect();
    var pos = {x: evt.clientX - rect.left, y: evt.clientY - rect.top};
    pressState.processMouseDown(me, pos);
  });
  this.canvas.addEventListener('mouseup', function(evt) {
    var rect = me.canvas.getBoundingClientRect();
    var pos = {x: evt.clientX - rect.left, y: evt.clientY - rect.top};
    pressState.processMouseUp(me, pos);
  });
  this.canvas.addEventListener('mousemove', function(evt) {
    var rect = me.canvas.getBoundingClientRect();
    var pos = {x: evt.clientX - rect.left, y: evt.clientY - rect.top};
    pressState.processMouseMove(me, pos);
  });

  //this.paint();
}

function AutomatonBBModelDisplay(canvasID, model) {
  this.canvas = document.getElementById(canvasID);
  this.context = this.canvas.getContext("2d");
  this.model = model;
  this.showStateName = false;

  this.paint = function() {
    this.context.clearRect(0, 0, this.canvas.width, this.canvas.height);
    this.context.fillStyle = "#ffffee";
    this.context.strokeStyle = "#aaaa00";
    this.context.font = 'bold 12px sans';
    this.context.textAlign = 'center';      
    this.context.textBaseline='middle';

    if(this.model.inputs.length==0 && this.model.outputs.length==0) return;

    let lvalMaxWidth = 0;
    let lnameMaxWidth = 0;
    for(let i = 0; i < this.model.inputs.length; i++) {
      let inp = this.model.inputs[i];
      let w = this.context.measureText(inp.name).width;
      if(lnameMaxWidth<w) lnameMaxWidth=w;
      for(let j = 0; j < inp.values.length; j++) {
        let w = this.context.measureText(inp.values[j]).width;
        if(lvalMaxWidth<w) lvalMaxWidth=w;
      }
    }
    let rvalMaxWidth = 0;
    let rnameMaxWidth = 0;
    for(let i = 0; i < this.model.outputs.length; i++) {
      let outp = this.model.outputs[i];
      let w = this.context.measureText(outp.name).width;
      if(rnameMaxWidth<w) rnameMaxWidth=w;
      for(let j = 0; j < outp.values.length; j++) {
        let w = this.context.measureText(outp.values[j]).width;
        if(rvalMaxWidth<w) rvalMaxWidth=w;
      }
    }

    let stnameMaxWidth = 0;
    for(let i = 0; i < this.model.stateNames.length; i++) {
      let w = this.context.measureText(this.model.stateNames[i]).width;
      if(stnameMaxWidth<w) stnameMaxWidth=w;
    }

    let innerMargin = 5;
    let outerMargin = 5;
    let connectorWidth = 10;

    let minimalBoxWidth = 200;
    let actualBoxWidth = lnameMaxWidth + innerMargin + stnameMaxWidth + innerMargin + rnameMaxWidth;
    if(actualBoxWidth<minimalBoxWidth) actualBoxWidth = minimalBoxWidth;

    let totalWidth = (lvalMaxWidth===0 ? 0 : lvalMaxWidth + outerMargin + connectorWidth) + actualBoxWidth + (rvalMaxWidth===0 ? 0 : rvalMaxWidth + outerMargin + connectorWidth);

    let dw = this.canvas.width - totalWidth;
    let ltx = dw/2 + (lvalMaxWidth===0 ? 0 : lvalMaxWidth + outerMargin + connectorWidth);
    let lty = 5;

    let ccHeight = 20;
    let minimalBoxHeight = 50;
    let actualBoxHeight = Math.max(this.model.outputs.length,this.model.inputs.length)*ccHeight;
    if(actualBoxHeight<minimalBoxHeight) actualBoxHeight=minimalBoxHeight;
    let pushyout = (actualBoxHeight-this.model.outputs.length*ccHeight)/2;
    let pushyin = (actualBoxHeight-this.model.inputs.length*ccHeight)/2;

    this.context.fillStyle = "#ffffee";
    this.context.strokeStyle = "#aaaa00";
    this.context.beginPath();
    this.context.rect(ltx,lty,actualBoxWidth,actualBoxHeight);
    this.context.fill();
    this.context.stroke();

    if(this.showStateName) {
      let st = this.model.findActiveState();
      if(st!=null) {
        this.context.textAlign = 'center';      
        this.context.textBaseline='middle';
        this.context.fillStyle = "#000000";
        this.context.fillText(st.stateName, ltx+actualBoxWidth/2, lty+actualBoxHeight/2);
      }
    }

    for(let i = 0; i < this.model.inputs.length; i++) {
      let inp = this.model.inputs[i];
      
      this.context.strokeStyle = "#000000";
      this.context.beginPath();
      this.context.moveTo(ltx, lty+pushyin+ccHeight/2 + i*ccHeight);
      this.context.lineTo(ltx-connectorWidth, lty+pushyin+ccHeight/2 + i*ccHeight);
      this.context.stroke();

      this.context.textAlign = 'left';      
      this.context.textBaseline='middle';
      this.context.fillStyle = "#000000";
      this.context.fillText(inp.name, ltx+outerMargin, lty+pushyin+ccHeight/2 + i*ccHeight);
      if(i<this.model.currentInput.length) {
        this.context.textAlign = 'right';      
        this.context.textBaseline='middle';
        this.context.fillText(this.model.currentInput[i], ltx-connectorWidth-outerMargin, lty+pushyin+ccHeight/2 + i*ccHeight);
      }
    }

    for(let i = 0; i < this.model.outputs.length; i++) {
      let outp = this.model.outputs[i];
      
      this.context.strokeStyle = "#000000";
      this.context.beginPath();
      this.context.moveTo(ltx+actualBoxWidth, lty+pushyout+ccHeight/2 + i*ccHeight);
      this.context.lineTo(ltx+actualBoxWidth+connectorWidth, lty+pushyout+ccHeight/2 + i*ccHeight);
      this.context.stroke();

      this.context.textAlign = 'right';      
      this.context.textBaseline='middle';
      this.context.fillStyle = "#000000";
      this.context.fillText(outp.name, ltx+actualBoxWidth-outerMargin, lty+pushyout+ccHeight/2 + i*ccHeight);
      if(i<this.model.currentOutput.length) {
        this.context.textAlign = 'left';      
        this.context.textBaseline='middle';
        this.context.fillText(this.model.currentOutput[i], ltx+actualBoxWidth+connectorWidth+outerMargin, lty+pushyout+ccHeight/2 + i*ccHeight);
      }
    }

/*      this.context.fillStyle = "#000000";
      this.context.font = 'bold 10px sans';
      this.context.textAlign = 'left';      
      this.context.textBaseline='top';
      this.context.fillText(str, 0, 0);
*/
  }
}

